import concurrent.futures
import re
import urllib.request

import stargazing.audio.audio_player as audio_ap
import stargazing.config.config as config
from stargazing.utils.helper_funcs import silent_stderr, start_daemon_thread


class AudioController():
    """Audio manager, pre-loads the audio players specified in the settings.json and allows audio players to be created and stream via YouTube search

    @param volume: Initial volume level."""

    def __init__(self, volume=100) -> None:
        self.saved_youtube_player_urls = config.get_saved_youtube_player_urls()
        self.loaded_players = {
            name: None for name in self.saved_youtube_player_urls}
        start_daemon_thread(target=self.__load_audio_players)

        self.playing = None
        self.playing_name = "offline"

        self.volume = volume

    def stop(self) -> None:
        if self.playing:
            self.playing.stop()

    def offline(self) -> None:
        self.stop()
        self.playing = None
        self.playing_name = "offline"

    def set_volume(self, vol: int) -> None:
        self.volume = vol
        if self.playing:
            self.playing.set_volume(vol)

    def get_volume(self) -> int:
        return self.volume

    def set_loaded_player(self, loaded_player_name: str) -> None:
        """Stops the current player, loads the given player name and closes the menu"""

        self.stop()

        # If player has not loaded (not enough time or error)
        if not self.loaded_players[loaded_player_name]:
            start_daemon_thread(target=self.set_youtube_player_from_url,
                                args=[self.saved_youtube_player_urls[loaded_player_name], loaded_player_name])
            # TODO: set self.loaded_players[loaded_player_name] with this player

        # If player has loaded
        else:
            self.playing = self.loaded_players[loaded_player_name]
            self.playing_name = loaded_player_name

            self.playing.set_volume(self.volume)
            self.playing.play()

    def set_youtube_player_from_url(self, youtube_url: str, player_name="") -> str:

        self.stop()

        self.playing_name = "loading audio..."
        self.playing = audio_ap.YoutubeAudioPlayer.safe_create(
            youtube_url, True)

        if self.playing:
            self.playing_name = self.playing.video_titles[0] if not player_name else player_name
            self.playing.set_volume(self.volume)
            self.playing.play()
        else:
            self.playing_name = "error loading audio"

    def set_youtube_player_from_query(self, search_query: str) -> str:

        self.stop()

        self.playing_name = "searching youtube..."

        search = search_query.replace(" ", "+")
        youtube_search = f"https://www.youtube.com/results?search_query={search}"

        html = urllib.request.urlopen(youtube_search)
        video_ids = re.findall(r"watch\?v=(\S{11})", html.read().decode())
        url = "https://www.youtube.com/watch?v=" + video_ids[0]

        self.set_youtube_player_from_url(url)

    def __load_audio_players(self) -> None:
        silent_yt_audio_init = silent_stderr(
            lambda url: audio_ap.YoutubeAudioPlayer.safe_create(url, True))

        with concurrent.futures.ThreadPoolExecutor() as exec:
            futures_to_name = {exec.submit(
                silent_yt_audio_init, url): name for name, url in self.saved_youtube_player_urls.items()}

            for future in concurrent.futures.as_completed(futures_to_name):
                name = futures_to_name[future]
                player = future.result()

                self.loaded_players[name] = player



from datetime import datetime
import math
import time
from typing import Tuple

from stargazing.utils.format_funcs import format_pomodoro_time


class Timer():

    def __init__(self, secs: int) -> None:
        self.interval = secs
        self.local_start_time = datetime.now()
        self.start_time = None
        self.paused_time = None
        self.elapsed_time = 0

    def start(self) -> None:
        self.start_time = time.time()
        self.elapsed_time = 0
        self.paused_time = None

    def continue_(self) -> None:
        paused_time_elapsed = time.time() - self.paused_time
        self.start_time += paused_time_elapsed
        self.paused_time = None

    def pause(self) -> None:
        self.paused_time = time.time()

    def update(self) -> Tuple[int, bool]:
        if not self.start_time:
            return 0, False

        old_elapsed_time = self.elapsed_time

        if self.paused_time:
            self.elapsed_time = self.paused_time - self.start_time
        else:
            self.elapsed_time = time.time() - self.start_time

        time_diff = self.elapsed_time - old_elapsed_time if old_elapsed_time else 0

        if self.elapsed_time > self.interval:
            return time_diff, True
        return time_diff, False

    @property
    def remaining_time(self) -> str:
        elapsed_time_secs = math.floor(self.elapsed_time)
        return format_pomodoro_time(self.interval - elapsed_time_secs)



import math
import random
import time
from typing import List
import os.path as path

DEFAULT_STARS_PATH = f"{path.dirname(path.abspath(__file__))}/../res/stars.txt"


class StarsGenerator():
    """Generates printable stars for the background of stargazing."""

    def __init__(self) -> None:
        self.stars = self.load_default_stars()
        self.last_printed_time = time.time()

        self.gen_time_interval = 5
        self.gen_random_threshold = 0.1
        self.gen_max_dist = 5

    def get_stars(self) -> List[str]:
        curr_time = time.time()

        if curr_time - self.last_printed_time > self.gen_time_interval:
            self.last_printed_time = curr_time
            self.stars = self.gen_new_stars()

        return self.stars

    def load_default_stars(self) -> List[str]:
        with open(DEFAULT_STARS_PATH, "r", encoding="utf-8") as f:
            lines = f.readlines()
            trimmed_lines = [line.rstrip("\n") for line in lines]
            return trimmed_lines

    def gen_new_stars(self) -> List[str]:
        new_stars = []

        for line in self.stars:
            chars = list(line)
            for i in range(len(chars)):
                if random.random() < self.gen_random_threshold:
                    swap_dist = math.ceil(random.random() * self.gen_max_dist)
                    swap_i = (i + swap_dist) % len(chars)
                    chars[i], chars[swap_i] = chars[swap_i], chars[i]

            new_stars.append("".join(chars))

        return new_stars



from collections import namedtuple
from datetime import datetime, timedelta
from typing import List, Union
import os.path as path

import stargazing.pomodoro.timer as pomo_t
import stargazing.project.project_controller as project_pc

PomodoroRecord = namedtuple(
    "PomodoroRecord", ["project_name", "start_time", "length"])

TIME_FORMAT = "%d/%m/%Y %H:%M:%S"

PROJECT_DATABASE_PATH = f"{path.dirname(path.abspath(__file__))}/../data/projects.txt"
POMODORO_DATABASE_PATH = f"{path.dirname(path.abspath(__file__))}/../data/pomodoros.txt"

DATABASE_DELIMITER_CHAR = "|"


def insert_project(project: project_pc.Project) -> bool:
    with open(PROJECT_DATABASE_PATH, "r") as file:
        lines = file.readlines()

    stripped_lines = [line.rstrip("\n") for line in lines]
    if project.name in stripped_lines:
        return False

    with open(PROJECT_DATABASE_PATH, "a") as file:
        file.write(f"{project.name}\n")

    return True


def get_all_projects() -> List[project_pc.Project]:
    with open(PROJECT_DATABASE_PATH, "r") as file:
        lines = file.readlines()

    stripped_lines = [line.rstrip("\n") for line in lines]
    projects = {name: project_pc.Project(name) for name in stripped_lines}

    pomo_records = get_all_pomodoros()
    current_time = datetime.now()

    for pomo_record in pomo_records:
        if pomo_record.project_name in projects:
            is_current_day = (pomo_record.start_time.year == current_time.year and pomo_record.start_time.month ==
                              current_time.month and pomo_record.start_time.day == current_time.day)
            projects[pomo_record.project_name].add_time(
                pomo_record.length, is_current_day)

    return list(projects.values())


def get_todays_total_time() -> int:
    pomo_records = get_all_pomodoros()
    current_time = datetime.now()
    todays_total_time = 0

    for pomo_record in pomo_records:
        is_current_day = (pomo_record.start_time.year == current_time.year and pomo_record.start_time.month ==
                          current_time.month and pomo_record.start_time.day == current_time.day)
        if is_current_day:
            todays_total_time += pomo_record.length

    return todays_total_time


def insert_pomodoro(project: project_pc.Project, timer: pomo_t.Timer) -> None:
    with open(POMODORO_DATABASE_PATH, "a") as file:
        start_time = timer.local_start_time.strftime(TIME_FORMAT)
        file.write(f"{project.name}|{start_time}|{timer.elapsed_time}\n")


def get_all_pomodoros() -> List[PomodoroRecord]:
    with open(POMODORO_DATABASE_PATH, "r") as file:
        lines = file.readlines()

    pomo_records = []
    for line in lines:
        project_name, start_time_str, length_str = line.split(
            DATABASE_DELIMITER_CHAR)

        start_time = datetime.strptime(start_time_str, TIME_FORMAT)
        length = float(length_str)

        pomo_records.append(PomodoroRecord(
            project_name, start_time, float(length)))

    return pomo_records


if __name__ == "__main__":
    print(get_todays_total_time())



def format_pomodoro_time(seconds: int, show_seconds=True) -> str:
    s = seconds
    m = s // 60
    s -= m * 60

    if not show_seconds:
        return m

    if m:
        format_s = '0' + str(s) if s < 10 else s
        return f"{m}:{format_s}"
    return s


def format_project_time(seconds: int) -> str:
    s = seconds
    h = s // 3600
    m = (s - h * 3600) // 60

    return f"{h}h {'0' + str(m) if m < 10 else m}m"



from blessed import Terminal
from functools import partial
from typing import Callable

import stargazing.pomodoro.pomodoro_controller as pomo_pc
from stargazing.utils.menu import Menu


class AutoStartMenu(Menu):
    """Menu interface to change the auto start option of the pomodoro timer.

    @param term: Instance of a Blessed terminal.
    @param on_close: Callback function to run when menu is closed.
    @param pomodoro_controller: Instance of a pomodoro controller."""

    def __init__(self, term: Terminal, on_close: Callable[[], None], pomodoro_controller: pomo_pc.PomodoroController) -> None:
        super().__init__(on_close, term.gray20_on_lavender)

        self.term = term
        self.pomodoro_controller = pomodoro_controller

        self.setup_menu()

    def set_autostart_and_close(self, option: bool) -> None:
        self.pomodoro_controller.autostart_setting = option
        super().handle_close()

    def setup_menu(self) -> None:
        for option in [True, False]:
            on_item_select = partial(
                self.set_autostart_and_close, option)
            index = super().add_item(str(option), on_item_select)

            if option == self.pomodoro_controller.autostart_setting:
                super().set_hover(index)



from blessed import Terminal
from functools import partial
from typing import Callable

import stargazing.config.config as config
import stargazing.pomodoro.pomodoro_controller as pomo_pc
from stargazing.utils.menu import Menu


class IntervalMenu(Menu):
    """Menu interface to change the interval settings of the pomodoro timer.

    @param term: Instance of a Blessed terminal.
    @param on_close: Callback function to run when menu is closed.
    @param pomo_controller: Instance of a pomodoro controller."""

    def __init__(self, term: Terminal, on_close: Callable[[], None], pomo_controller: pomo_pc.PomodoroController) -> None:
        super().__init__(on_close, term.gray20_on_lavender)

        self.term = term
        self.pomo_controller = pomo_controller

        self.interval_settings = self.__load_interval_settings()

        self.setup_menu()

    def __load_interval_settings(self) -> None:
        interval_times = config.get_interval_times()
        settings = [pomo_pc.PomodoroIntervalSettings(
            *interval_time) for interval_time in interval_times]
        return settings

    def set_interval_settings_and_close(self, interval_settings: pomo_pc.PomodoroIntervalSettings) -> None:
        self.pomo_controller.set_interval_settings(interval_settings)
        super().handle_close()

    def setup_menu(self) -> None:
        for interval_setting in self.interval_settings:
            on_item_select = partial(
                self.set_interval_settings_and_close, interval_setting)
            index = super().add_item(interval_setting.name, on_item_select)

            if interval_setting == self.pomo_controller.interval_settings:
                super().set_hover(index)



from blessed import Terminal
from functools import partial
from typing import Callable

import stargazing.audio.audio_controller as audio_ac
from stargazing.utils.menu import Menu


class VolumeMenu(Menu):
    """Menu interface to change the volume of the audio controller.

    @param term: Instance of a Blessed terminal.
    @param on_close: Callback function to run when menu is closed.
    @param audio_controller: Instance of an audio controller."""

    def __init__(self, term: Terminal, on_close: Callable[[], None], audio_controller: audio_ac.AudioController) -> None:
        super().__init__(on_close, term.gray20_on_lavender)

        self.term = term
        self.audio_controller = audio_controller

        self.setup_menu()

    def set_volume_and_close(self, option: bool) -> None:
        self.audio_controller.set_volume(option)
        super().handle_close()

    def setup_menu(self) -> None:
        for option in range(100, -1, -10):
            on_item_select = partial(
                self.set_volume_and_close, option)
            index = super().add_item(str(option), on_item_select)

            if option == self.audio_controller.volume:
                super().set_hover(index)



import logging
import logging.config
import os.path as path

LOGGING_CONFIG = {
    "version": 1,
    "formatters": {
        "simple": {
            "format": "%(asctime)s - %(levelname)s - %(message)s"
        },
        "detail": {
            "format": "%(asctime)s - %(module)s - %(funcName)s - %(levelname)s - %(message)s"
        }
    },
    "handlers": {
        "debugfile": {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "filename": f"{path.dirname(path.abspath(__file__))}/../logs/debug.log",
            "encoding": "utf8"
        },
        "errorfile": {
            "class": "logging.FileHandler",
            "level": "ERROR",
            "formatter": "detail",
            "filename": f"{path.dirname(path.abspath(__file__))}/../logs/error.log",
            "encoding": "utf8"
        }
    },
    "loggers": {
        "basic": {
            "level": "DEBUG",
            "handlers": [
                "debugfile", "errorfile"
            ],
            "propagate": False
        }
    }
}

logging.config.dictConfig(LOGGING_CONFIG)

logger = logging.getLogger("basic")



from __future__ import annotations
from typing import List, Tuple
import ujson
import os.path as path

import stargazing.pomodoro.pomodoro_controller as pomo_pc

CONFIG_FILE_PATH = f"{path.dirname(path.abspath(__file__))}/../config/settings.json"


def get_saved_youtube_player_urls() -> List[str]:
    with open(CONFIG_FILE_PATH) as file:
        data = ujson.load(file)
        return data["saved_youtube_player_urls"]


def get_interval_times() -> List[List[int]]:
    with open(CONFIG_FILE_PATH) as file:
        data = ujson.load(file)
        return data["interval_times"]


def get_last_session_data() -> Tuple[str, pomo_pc.PomodoroIntervalSettings, bool, int]:
    with open(CONFIG_FILE_PATH) as file:
        data = ujson.load(file)
        return (data["last_project_name"], pomo_pc.PomodoroIntervalSettings(*data["last_interval_time"]),
                data["last_autostart"], data["last_volume"])


def update_last_session_data(project_name: str, interval_settings: pomo_pc.PomodoroIntervalSettings,
                             autostart: bool, volume: int) -> None:
    with open(CONFIG_FILE_PATH, 'r+') as file:
        data = ujson.load(file)

        data["last_project_name"] = project_name
        data["last_interval_time"] = [
            interval_settings.work_secs, interval_settings.break_secs]
        data["last_autostart"] = autostart
        data["last_volume"] = volume

        file.seek(0)
        ujson.dump(data, file, indent=4)
        file.truncate()



from blessed import Terminal
from functools import partial
from typing import Callable

import stargazing.data.database as database
import stargazing.project.project_controller as proj_pc
from stargazing.utils.menu import Menu


class ProjectMenu(Menu):
    """Menu interface to create and switch between projects.

    @param term: Instance of a Blessed terminal.
    @param on_close: Callback function to run when menu is closed.
    @param pomodoro_controller: Instance of a project controller."""

    def __init__(self, term: Terminal, on_close: Callable[[], None], project_controller: proj_pc.ProjectController) -> None:
        super().__init__(on_close, term.gray20_on_lavender)

        self.term = term
        self.project_controller = project_controller

        self.create_new_project_mode = False
        self.create_new_project_name = ""

        self.setup_menu()

    def set_current_project_and_close(self, project: proj_pc.Project) -> None:
        self.project_controller.set_current_project(project)
        super().handle_close()

    def start_create_new_project_mode(self) -> None:
        self.create_new_project_mode = True
        super().replace_item(-1,
                             "> enter project name", self.finish_create_new_project_mode)

    def update_create_new_project_name(self, name: str) -> None:
        self.create_new_project_name = name
        super().replace_item(-1, "> " + self.create_new_project_name +
                             self.term.lightsteelblue1("█"), self.finish_create_new_project_mode)

    def cancel_create_new_project_mode(self) -> None:
        self.create_new_project_name = ""
        self.finish_create_new_project_mode()

    def finish_create_new_project_mode(self) -> None:
        self.create_new_project_name = self.create_new_project_name.strip()

        created_project = self.project_controller.create_and_insert_new_project(
            self.create_new_project_name)

        if created_project:
            on_item_select = partial(
                self.set_current_project_and_close, created_project)
            super().add_item(created_project.name, on_item_select, -1)

        self.create_new_project_name = ""
        self.create_new_project_mode = False
        super().replace_item(-1, self.term.underline("create new project"),
                             self.start_create_new_project_mode)

    def setup_menu(self) -> None:
        for project in self.project_controller.projects:
            on_item_select = partial(
                self.set_current_project_and_close, project)
            index = super().add_item(project.name, on_item_select)

            if project == self.project_controller.current:
                super().set_hover(index)

        super().add_item(self.term.underline("create new project"),
                         self.start_create_new_project_mode)

    def handle_key_up(self) -> None:
        if self.create_new_project_mode:
            return

        super().handle_key_up()

    def handle_key_down(self) -> None:
        if self.create_new_project_mode:
            return

        super().handle_key_down()

    def handle_key_escape(self) -> None:
        if not self.create_new_project_mode:
            super().handle_key_escape()
            return

        self.cancel_create_new_project_mode()

    def handle_key_backspace(self) -> None:
        if not self.create_new_project_mode or not self.create_new_project_name:
            return

        new_name = self.create_new_project_name[:-1]
        self.update_create_new_project_name(new_name)

    def handle_char_input(self, char: str) -> None:
        if not self.create_new_project_mode:
            super().handle_char_input(char)
            return

        if char == database.DATABASE_DELIMITER_CHAR:
            return

        new_name = self.create_new_project_name + char
        self.update_create_new_project_name(new_name)



from blessed import Terminal
from functools import partial
import sys
import os.path as path

import stargazing.audio.audio_controller as audio_ac
import stargazing.audio.player_menu as audio_pm
import stargazing.audio.volume_menu as audio_vm

import stargazing.config.config as config

import stargazing.pomodoro.autostart_menu as pomo_am
import stargazing.pomodoro.interval_menu as pomo_im
import stargazing.pomodoro.pomodoro_controller as pomo_pc
import stargazing.pomodoro.status_menu as pomo_sm

import stargazing.project.project_controller as proj_pc
import stargazing.project.project_menu as proj_pm

from stargazing.utils.logger import logger
from stargazing.utils.menu import Menu
import stargazing.utils.print_funcs as print_funcs
from stargazing.utils.stars import StarsGenerator


GAZING_PATH = f"{path.dirname(path.abspath(__file__))}/res/gazing.txt"


class Stargazing(Menu):

    def __init__(self) -> None:

        self.term = Terminal()

        super().__init__(on_close=self.handle_close, hover_dec=self.term.gray20_on_lavender)

        last_project_name, last_interval_settings, last_autostart, last_volume = config.get_last_session_data()

        self.audio_controller = audio_ac.AudioController(last_volume)
        self.project_controller = proj_pc.ProjectController(last_project_name)
        self.pomodoro_controller = pomo_pc.PomodoroController(
            self.project_controller, self.audio_controller, last_interval_settings, last_autostart)

        self.project_menu = proj_pm.ProjectMenu(
            self.term, self.close_submenu, self.project_controller)

        self.interval_menu = pomo_im.IntervalMenu(
            self.term, partial(self.close_submenu, True), self.pomodoro_controller)
        self.autostart_menu = pomo_am.AutoStartMenu(
            self.term, self.close_submenu, self.pomodoro_controller)
        self.status_menu = pomo_sm.StatusMenu(
            self.term, partial(self.close_submenu, True), self.pomodoro_controller)

        self.player_menu = audio_pm.PlayerMenu(
            self.term, self.close_submenu, self.audio_controller)
        self.volume_menu = audio_vm.VolumeMenu(
            self.term, self.close_submenu, self.audio_controller)

        self.menu = None
        self.submenu = None
        self.focused_menu = self

        self.running = False
        self.refresh_speed = 0.2

        self.stars_generator = StarsGenerator()
        self.setup_menu()

    def start(self) -> None:
        self.running = True

        with self.term.fullscreen(), self.term.cbreak(), self.term.hidden_cursor():
            print(self.term.home + self.term.clear)

            self.print_logo()
            self.print_gazing()
            self.print_stars()

            while self.running:
                self.print_stars()
                self.print_menu()
                self.print_submenu()

                sys.stdout.flush()

                inp = self.term.inkey(self.refresh_speed)

                if inp.is_sequence and inp.name == "KEY_UP":
                    self.focused_menu.handle_key_up()
                elif inp.is_sequence and inp.name == "KEY_DOWN":
                    self.focused_menu.handle_key_down()
                elif inp.is_sequence and inp.name == "KEY_ENTER":
                    self.focused_menu.handle_key_enter()
                elif inp.is_sequence and inp.name == "KEY_ESCAPE":
                    self.focused_menu.handle_key_escape()
                elif inp.is_sequence and inp.name == "KEY_BACKSPACE":
                    self.focused_menu.handle_key_backspace()
                elif inp and not inp.is_sequence:
                    self.focused_menu.handle_char_input(inp)

                self.pomodoro_controller.update_timer()

        self.__save_last_session_data()
        print("Exiting stargazing...")

    def handle_close(self) -> None:
        self.pomodoro_controller.finish_timer(disable_sound=True)
        self.running = False

    def handle_char_input(self, char) -> None:
        if char.lower() == "r":
            print(self.term.home + self.term.clear)

            self.print_logo()
            self.print_gazing()
            self.print_stars()
        else:
            super().handle_char_input(char)

    def open_submenu(self, submenu: Menu) -> None:
        self.submenu = submenu
        self.focused_menu = self.submenu

    def close_submenu(self, update_timer=False) -> None:
        self.print_stars()

        if update_timer:
            self.pomodoro_controller.update_timer()

        self.submenu = None
        self.focused_menu = self

    def toggle_pomodoro_display(self) -> None:
        self.pomodoro_controller.toggle_start_stop()
        self.pomodoro_controller.update_timer()

    def setup_menu(self) -> None:

        super().add_item(
            lambda: f"{self.term.bold('project')}: {self.term.lightcoral(self.project_controller.current.name)}",
            partial(self.open_submenu, self.project_menu))
        super().add_item(
            lambda: f"{self.term.bold('todays time')}: {self.term.lightskyblue3(self.project_controller.current.formatted_todays_time + ' | ' + self.project_controller.formatted_todays_total_time)}")
        super().add_item(
            lambda: f"{self.term.bold('total time')}: {self.term.lightskyblue3(self.project_controller.current.formatted_total_time)}")

        super().add_divider()

        super().add_item(
            lambda: f"{self.term.bold('pomodoro')}: {self.pomodoro_controller.interval_settings.name}",
            partial(self.open_submenu, self.interval_menu))
        super().add_item(
            lambda: f"{self.term.bold('auto-start')}: {self.pomodoro_controller.autostart_setting}",
            partial(self.open_submenu, self.autostart_menu))
        super().add_item(
            lambda: f"{self.term.bold('status')}: {self.pomodoro_controller.status.value}",
            partial(self.open_submenu, self.status_menu))

        super().add_divider()

        super().add_item(
            lambda: f"{self.term.bold('playing')}: {self.audio_controller.playing_name}",
            partial(self.open_submenu, self.player_menu))
        super().add_item(
            lambda: f"{self.term.bold('volume')}: {self.audio_controller.volume}",
            partial(self.open_submenu, self.volume_menu))

        super().add_divider()

        super().add_item(
            lambda: f"{self.pomodoro_controller.timer_display}",
            self.toggle_pomodoro_display)

    def __save_last_session_data(self) -> None:
        config.update_last_session_data(self.project_controller.current.name, self.pomodoro_controller.interval_settings,
                                        self.pomodoro_controller.autostart_setting, self.audio_controller.volume)

    # ========================================================
    # Terminal printing methods
    # ========================================================

    def print_menu(self) -> None:
        lines = super().get_print_strings()
        menu_lines, pomo_line = lines[:-1], lines[-1]

        mx, my = 6, 3
        print_funcs.print_lines_xy(self.term, mx, my, menu_lines,
                                   flush=False, max_width=30, trim=True)

        px, py = round(self.term.width / 2), self.term.height - 4
        print_funcs.print_xy(self.term, px, py, pomo_line,
                             flush=False, max_width=19, center=True)

    def print_submenu(self) -> None:
        if self.submenu is None:
            return

        lines = self.submenu.get_print_strings()
        x, y = 40, 3

        if self.submenu is self.project_menu:
            max_width = 30
        elif self.submenu is self.interval_menu:
            max_width = 7
        elif self.submenu is self.autostart_menu:
            max_width = 5
        elif self.submenu is self.status_menu:
            max_width = 12
        elif self.submenu is self.player_menu:
            max_width = 30
        elif self.submenu is self.volume_menu:
            max_width = 3

        print_funcs.print_lines_xy(self.term, x, y, lines,
                                   flush=False, max_width=max_width, trim=True)

    def print_logo(self) -> None:
        print_funcs.print_xy(self.term, 0, 0,
                             self.term.gray20_on_white(self.term.bold(' ' + self.term.link('https://github.com/mtu2/stargazing', 'stargazing') + ' ')))

    def print_gazing(self) -> None:
        with open(GAZING_PATH, "r", encoding="utf-8") as f:
            lines = f.readlines()
            dec_lines = [self.term.aliceblue(line) for line in lines]

            x, y = 1, self.term.height - len(lines)

            print_funcs.print_lines_xy(self.term, x, y, dec_lines)

    def print_stars(self) -> None:
        stars = self.stars_generator.get_stars()
        dec_lines = [self.term.aliceblue(line) for line in stars]

        x, y = 2, 1

        print_funcs.print_lines_xy(self.term, x, y, dec_lines, flush=False)


def run_stargazing():
    """Main entry point for script"""
    stargazing = Stargazing()
    stargazing.start()


if __name__ == "__main__":
    run_stargazing()



from blessed import Terminal
from functools import partial
from typing import Callable

import stargazing.audio.audio_controller as audio_ac
from stargazing.utils.helper_funcs import start_daemon_thread
from stargazing.utils.menu import Menu


class PlayerMenu(Menu):
    """Menu interface to change the current player in the audio controller.

    @param term: Instance of a Blessed terminal.
    @param on_close: Callback function to run when menu is closed.
    @param audio_controller: Instance of an audio controller."""

    def __init__(self, term: Terminal, on_close: Callable[[], None], audio_controller: audio_ac.AudioController) -> None:
        super().__init__(on_close, term.gray20_on_lavender)

        self.term = term
        self.audio_controller = audio_controller

        self.search_youtube_mode = False
        self.search_youtube_query = ""

        self.setup_menu()

    def set_offline_and_close(self) -> None:
        super().handle_close()
        self.audio_controller.offline()

    def set_loaded_player_and_close(self, loaded_player_name: str) -> None:
        """Stops the current player, loads the given player name and closes the menu"""
        super().handle_close()
        self.audio_controller.set_loaded_player(loaded_player_name)

    def set_youtube_player_from_url_and_close(self, player_name: str, youtube_url: str) -> str:
        super().handle_close()
        self.audio_controller.set_youtube_player_from_url(
            player_name, youtube_url)

    def set_youtube_player_from_query_and_close(self, search_query: str) -> str:
        super().handle_close()
        self.audio_controller.set_youtube_player_from_query(search_query)

    def start_search_youtube_mode(self):
        self.search_youtube_mode = True
        super().replace_item(-1,
                             "> search youtube", self.finish_search_youtube_mode)

    def update_search_youtube_query(self, query):
        self.search_youtube_query = query
        super().replace_item(-1, "> " + self.search_youtube_query +
                             self.term.lightsteelblue1("█"), self.finish_search_youtube_mode)

    def cancel_search_youtube_mode(self):
        self.search_youtube_query = ""
        self.finish_search_youtube_mode()

    def finish_search_youtube_mode(self):
        self.search_youtube_query = self.search_youtube_query.strip()

        if self.search_youtube_query:
            start_daemon_thread(target=self.set_youtube_player_from_query_and_close, args=(
                self.search_youtube_query,))

        self.search_youtube_query = ""
        self.search_youtube_mode = False
        super().replace_item(-1,
                             self.term.underline("search youtube"), self.start_search_youtube_mode)

    def setup_menu(self) -> None:
        super().add_item("offline", self.set_offline_and_close)

        for loaded_player_name in self.audio_controller.loaded_players:
            on_item_select = partial(
                self.set_loaded_player_and_close, loaded_player_name)
            super().add_item(loaded_player_name, on_item_select)

        super().add_item(self.term.underline("search youtube"),
                         self.start_search_youtube_mode)
        super().set_hover(0)

    def handle_key_up(self) -> None:
        if self.search_youtube_mode:
            return

        super().handle_key_up()

    def handle_key_down(self) -> None:
        if self.search_youtube_mode:
            return

        super().handle_key_down()

    def handle_key_escape(self) -> None:
        if not self.search_youtube_mode:
            super().handle_key_escape()
            return

        self.cancel_search_youtube_mode()

    def handle_key_backspace(self) -> None:
        if not self.search_youtube_mode or not self.search_youtube_query:
            return

        new_query = self.search_youtube_query[:-1]
        self.update_search_youtube_query(new_query)

    def handle_char_input(self, char: str) -> None:
        if not self.search_youtube_mode:
            super().handle_char_input(char)
            return

        new_query = self.search_youtube_query + char
        self.update_search_youtube_query(new_query)



from typing import Callable, Union

from stargazing.utils.helper_funcs import check_null_fn


class MenuItem():
    """Item for the menu user interface.

    @param text: Display text for the item, either a string or a function that returns a string
    @param handle_select: Callback function when the item is selected"""

    def __init__(self, text: Union[str, Callable[[], str]], handle_select: Callable[[], None] = None):
        self._text = text
        self.handle_select = check_null_fn(handle_select)

    @property
    def text(self) -> str:
        if isinstance(self._text, str):
            return self._text
        return self._text()


class Menu():
    """Menu user interface - all user interaction occurs through a main menu and several submenus.

    @param on_close: Callback function to run when menu is closed.
    @param hover_dec: Blessed terminal function to provide colour for the currently hovered item."""

    def __init__(self, on_close: Callable[[bool], None] = None, hover_dec: Callable[[str], str] = None) -> None:

        # List of menu items ordered top to bottom in appearance
        self.items = []

        # List of indexes for dividers - appears after the index of the corresponding menu item
        self.dividers = []

        # Index of currently hovered item
        self.hover_index = 0

        # Callback function on menu close
        self.on_close = check_null_fn(on_close)

        # Function to decorate currently hovered item
        self.hover_dec = check_null_fn(hover_dec)

    # ========================================================
    # Create menu methods
    # ========================================================

    def add_item(self, text: Union[str, Callable[[], str]], handle_item_select: Callable[[], None] = None,
                 index: int = None) -> int:
        """Creates and adds a menu item at the given index. If no index is given, the item 
        is added to the bottom of the menu. Returns the index of the added item."""

        if index is None:
            index = len(self.items)

        if index < -len(self.items) or index > len(self.items):
            raise IndexError(f"Entered invalid index: {index}")

        item = MenuItem(text, handle_item_select)
        self.items.insert(index, item)

        return index

    def replace_item(self, index: int, text: Union[str, Callable[[], str]],
                     handle_item_select: Callable[[], None] = None) -> None:
        """Creates and replaces a menu item at the given index."""

        if index < -len(self.items) or index >= len(self.items):
            raise IndexError(f"Entered invalid index: {index}")

        item = MenuItem(text, handle_item_select)
        self.items[index] = item

    def add_divider(self) -> int:
        """Creates and adds a divider (a blank menu item) to the bottom of the menu.
        Returns the index of the added divider."""

        self.dividers.append(len(self.items) - 1)

        return len(self.dividers) - 1

    # ========================================================
    # User interaction handlers
    # ========================================================

    def handle_key_up(self) -> None:
        """Moves the hover index up to the next menu item."""

        new_hover_index = max(self.hover_index - 1, 0)
        self.hover_index = new_hover_index

    def handle_key_down(self) -> None:
        """Moves the hover index down to the next menu item."""

        new_hover_index = min(self.hover_index + 1, len(self.items) - 1)
        self.hover_index = new_hover_index

    def handle_key_enter(self) -> None:
        """Selects the current hovered menu item and closes the menu (if function is given)"""

        self.items[self.hover_index].handle_select()

    def handle_key_escape(self) -> None:
        """Handles escape key sequences"""
        self.handle_close()

    def handle_key_backspace(self) -> None:
        """Handles backspace key sequences"""
        pass

    def handle_char_input(self, char) -> None:
        """Handles character keyboard input"""
        if char.lower() == "q":
            self.handle_close()

    def handle_close(self) -> None:
        """Handles menu close"""
        self.on_close()

    # ========================================================
    # Manual interaction handlers
    # ========================================================

    def set_hover(self, index: int) -> None:
        """Manually sets the hover index"""

        if index < 0 or index >= len(self.items):
            raise IndexError

        self.hover_index = index

    # ========================================================
    # Terminal printing helper methods
    # ========================================================

    def get_print_strings(self) -> str:
        """Returns a list of strings formatted for terminal printing"""

        print_strings = []
        dividers_p = 0

        for i, item in enumerate(self.items):
            while dividers_p < len(self.dividers) and self.dividers[dividers_p] < i:
                print_strings.append("")
                dividers_p += 1

            if self.hover_index == i:
                print_strings.append(self.hover_dec(item.text))
            else:
                print_strings.append(item.text)

        return print_strings



from blessed import Terminal
from typing import Callable

import stargazing.pomodoro.pomodoro_controller as pomo_pc
from stargazing.utils.menu import Menu


class StatusMenu(Menu):
    """Menu interface to manually change the status of the pomodoro timer.

    @param term: Instance of a Blessed terminal.
    @param on_close: Callback function to run when menu is closed.
    @param pomodoro_controller: Instance of a pomodoro controller."""

    def __init__(self, term: Terminal, on_close: Callable[[], None], pomodoro_controller: pomo_pc.PomodoroController) -> None:
        super().__init__(on_close, term.gray20_on_lavender)

        self.term = term
        self.pomodoro_controller = pomodoro_controller

        self.setup_menu()

    def finish_timer_and_close(self) -> None:
        self.pomodoro_controller.finish_timer()
        super().handle_close()

    def reset_timer_and_close(self) -> None:
        self.pomodoro_controller.reset_timer()
        super().handle_close()

    def setup_menu(self) -> None:
        super().add_item("finish timer", self.finish_timer_and_close)
        super().add_item("reset timer", self.reset_timer_and_close)



from __future__ import annotations
from enum import Enum
from typing import List
import os.path as path

import stargazing.data.database as database
import stargazing.audio.audio_controller as audio_ac
import stargazing.audio.audio_player as audio_ap
import stargazing.pomodoro.timer as pomo_t
import stargazing.project.project_controller as proj_pc
from stargazing.utils.format_funcs import format_pomodoro_time

ALARM_START_PATH = f"{path.dirname(path.abspath(__file__))}/../res/alarm_start.mp3"
ALARM_FINISH_PATH = f"{path.dirname(path.abspath(__file__))}/../res/alarm_finish.mp3"


class PomodoroIntervalSettings():
    """Interval settings for the pomodoro timer.

    @param work_secs: Number of seconds for the work interval of the timer.
    @param break_secs: Number of seconds for the break interval of the timer."""

    def __init__(self, work_secs: int, break_secs: int) -> None:
        self.work_secs = work_secs
        self.break_secs = break_secs

    @property
    def name(self) -> str:
        return f"{format_pomodoro_time(self.work_secs, False)} + {format_pomodoro_time(self.break_secs, False)}"

    def __eq__(self, o: PomodoroIntervalSettings) -> bool:
        return self.work_secs == o.work_secs and self.break_secs == o.break_secs

    def __ne__(self, o: PomodoroIntervalSettings) -> bool:
        return not self.__eq__(o)


class PomodoroStatus(Enum):
    INACTIVE = "inactive"
    WORK = "work"
    BREAK = "break"
    PAUSED_WORK = "paused work"
    PAUSED_BREAK = "paused break"
    FINISHED_WORK = "finished work"
    FINISHED_BREAK = "finished break"


class PomodoroController():
    """Pomodoro manager, containing current pomodoro timer, status, autostart option and interval settings.

    @param project_controller: Instance of a project controller.
    @param audio_controller: Instance of an audio controller."""

    def __init__(self, project_controller: proj_pc.ProjectController, audio_controller: audio_ac.AudioController,
                 interval_time: PomodoroIntervalSettings = None, last_autostart=True) -> None:

        self.project_controller = project_controller
        self.audio_controller = audio_controller

        self.interval_settings = interval_time if interval_time else PomodoroIntervalSettings(
            2400, 600)
        self.autostart_setting = last_autostart

        self.timer = pomo_t.Timer(self.interval_settings.work_secs)
        self.status = PomodoroStatus.INACTIVE

    def finish_timer(self, disable_sound=False) -> None:
        if self.status in (PomodoroStatus.WORK, PomodoroStatus.PAUSED_WORK):
            database.insert_pomodoro(
                self.project_controller.current, self.timer)
            self.timer = pomo_t.Timer(self.interval_settings.break_secs)

            if not disable_sound:
                self.__play_alarm_sound(ALARM_FINISH_PATH)

            if self.autostart_setting:
                self.timer.start()
                self.status = PomodoroStatus.BREAK
            else:
                self.status = PomodoroStatus.FINISHED_WORK
        elif self.status in (PomodoroStatus.BREAK, PomodoroStatus.PAUSED_BREAK):
            self.timer = pomo_t.Timer(self.interval_settings.work_secs)

            if self.autostart_setting:
                self.timer.start()
                self.status = PomodoroStatus.WORK

                if not disable_sound:
                    self.__play_alarm_sound(ALARM_START_PATH)
            else:
                self.status = PomodoroStatus.FINISHED_BREAK

    def reset_timer(self) -> None:
        if self.status in (PomodoroStatus.WORK, PomodoroStatus.PAUSED_WORK, PomodoroStatus.FINISHED_WORK):
            database.insert_pomodoro(
                self.project_controller.current, self.timer)
            self.timer = pomo_t.Timer(self.interval_settings.work_secs)

            self.timer.start()
            self.status = PomodoroStatus.WORK
        elif self.status in (PomodoroStatus.BREAK, PomodoroStatus.PAUSED_BREAK, PomodoroStatus.FINISHED_BREAK):
            self.timer = pomo_t.Timer(self.interval_settings.break_secs)

            self.timer.start()
            self.status = PomodoroStatus.BREAK

    def update_timer(self) -> None:

        time_diff, timer_complete = self.timer.update()

        if self.status == PomodoroStatus.WORK:
            self.project_controller.add_todays_total_time(time_diff)
            self.project_controller.current.add_time(time_diff, True)

        if timer_complete:
            self.finish_timer()

    def toggle_start_stop(self) -> None:
        if self.status in (PomodoroStatus.INACTIVE, PomodoroStatus.FINISHED_BREAK):
            self.timer.start()
            self.status = PomodoroStatus.WORK

            self.__play_alarm_sound(ALARM_START_PATH)

        elif self.status == PomodoroStatus.PAUSED_WORK:
            self.timer.continue_()
            self.status = PomodoroStatus.WORK

        elif self.status == PomodoroStatus.FINISHED_WORK:
            self.timer.start()
            self.status = PomodoroStatus.BREAK

        elif self.status == PomodoroStatus.PAUSED_BREAK:
            self.timer.continue_()
            self.status = PomodoroStatus.BREAK

        elif self.status == PomodoroStatus.WORK:
            self.timer.pause()
            self.status = PomodoroStatus.PAUSED_WORK
        elif self.status == PomodoroStatus.BREAK:
            self.timer.pause()
            self.status = PomodoroStatus.PAUSED_BREAK

    def set_interval_settings(self, interval_settings: PomodoroIntervalSettings) -> None:
        self.interval_settings = interval_settings

        # Edit current timer settings without resetting
        if self.status in (PomodoroStatus.INACTIVE, PomodoroStatus.WORK, PomodoroStatus.PAUSED_WORK):
            self.timer.interval = interval_settings.work_secs
        else:
            self.timer.interval = interval_settings.break_secs

    def __play_alarm_sound(self, path) -> None:
        curr_vol = self.audio_controller.get_volume()
        audio_decr = 15

        self.audio_controller.set_volume(max(curr_vol - audio_decr, 0))

        alarm = audio_ap.AudioPlayer(path)
        alarm.set_volume(curr_vol)
        alarm.play()

        # TODO: this needs to be async - wait for the alarm length
        self.audio_controller.set_volume(curr_vol)

    @property
    def timer_display(self) -> str:
        if self.status in (PomodoroStatus.INACTIVE, PomodoroStatus.FINISHED_BREAK):
            return "START TIMER"
        elif self.status == PomodoroStatus.WORK:
            return f"BREAK IN {self.timer.remaining_time}"
        elif self.status == PomodoroStatus.BREAK:
            return f"POMODORO IN {self.timer.remaining_time}"
        elif self.status == PomodoroStatus.PAUSED_WORK:
            return f"PAUSED [WORK {self.timer.remaining_time}]"
        elif self.status == PomodoroStatus.PAUSED_BREAK:
            return f"PAUSED [BREAK {self.timer.remaining_time}]"
        elif self.status == PomodoroStatus.FINISHED_WORK:
            return "START BREAK"



from __future__ import annotations  # allow typing of own class inside its class
from typing import Iterable, Union

import pafy
import pafy.backend_youtube_dl  # prevent lazy importing
import vlc

from stargazing.utils.helper_funcs import check_iterable
from stargazing.utils.logger import logger


class AudioPlayer():

    def __init__(self, sources: Union[str, Iterable[str]], loop=False):
        self.vlc_instance = vlc.Instance()
        self.vlc_instance.log_unset()

        self.player = self.vlc_instance.media_list_player_new()
        self.media_list = self.create_media_list(check_iterable(sources))

        if loop:
            self.player.set_playback_mode(vlc.PlaybackMode.loop)

        self.player.set_media_list(self.media_list)

    def create_media_list(self, sources: Iterable[str]) -> vlc.MediaList:
        media_list = self.vlc_instance.media_list_new()

        for source in sources:
            media = self.vlc_instance.media_new(source)
            media.add_option(":no-video")
            media_list.add_media(media)

        return media_list

    def play(self) -> None:
        self.player.play()

    def pause(self) -> None:
        self.player.pause()

    def stop(self) -> None:
        self.player.stop()

    def set_volume(self, vol: int) -> None:
        self.player.get_media_player().audio_set_volume(vol)

    def get_volume(self) -> int:
        return self.player.get_media_player().audio_get_volume()


class YoutubeAudioPlayer(AudioPlayer):

    def __init__(self, youtube_urls: Union[str, Iterable[str]], loop=False) -> None:

        self.video_titles, self.playurls = self.__get_youtube_titles_and_playurls(
            youtube_urls)
        super().__init__(self.playurls, loop)

    def __get_youtube_titles_and_playurls(self, youtube_urls: Union[str, Iterable[str]]):
        titles = []
        playurls = []

        for youtube_url in check_iterable(youtube_urls):
            video = pafy.new(youtube_url)
            titles.append(video.title)

            best = video.getbest()
            playurl = best.url
            playurls.append(playurl)

        return titles, playurls

    @staticmethod
    def safe_create(youtube_urls: Union[str, Iterable[str]], loop=False) -> Union[None, YoutubeAudioPlayer]:
        try:
            yt_audio_player = YoutubeAudioPlayer(youtube_urls, loop)
            return yt_audio_player
        except Exception as e:
            logger.error(
                f"Failed to create YoutubeAudioPlayer with inputs: {youtube_urls}, {loop}. Full message: {e}")
            return None



from __future__ import annotations
import math
from typing import List, Union

import stargazing.data.database as database
from stargazing.utils.format_funcs import format_project_time


class Project():
    """A named project counting total pomodoro time.

    @param name: Display name for the project"""

    def __init__(self, name: str) -> None:
        self.name = name
        self.todays_time = 0
        self.total_time = 0

    def add_time(self, secs: float, today_time=False) -> None:
        self.total_time += secs
        if today_time:
            self.todays_time += secs

    @property
    def formatted_todays_time(self) -> str:
        return format_project_time(math.floor(self.todays_time))

    @property
    def formatted_total_time(self) -> str:
        return format_project_time(math.floor(self.total_time))

    def __str__(self) -> str:
        return f"{self.name} | Today: {self.todays_time}secs, Total: {self.total_time}secs"

    def __eq__(self, o: Project) -> bool:
        return self.name == o.name

    def __ne__(self, o: Project) -> bool:
        return not self.__eq__(o)


class ProjectController():
    """Project manager, contains the currently loaded project and allows new projects to be created.

    @param project_name: Name of the default project to be loaded."""

    def __init__(self, project_name="") -> None:
        self.projects = self.__load_projects()
        self.current = self.projects[0]
        if project_name:
            for project in self.projects:
                if project.name == project_name:
                    self.current = project
                    break
        self.todays_total_time = self.__load_todays_total_time()

    def set_current_project(self, project: Project) -> None:
        self.current = project

    def create_and_insert_new_project(self, project_name: str) -> Union[Project, None]:
        project_name = project_name.strip()

        if not project_name:
            return None

        new_project = Project(project_name)
        insert_successful = database.insert_project(new_project)

        if not insert_successful:
            return None

        self.projects.append(new_project)
        return new_project

    def add_todays_total_time(self, secs: float) -> None:
        self.todays_total_time += secs

    def __load_projects(self) -> List[Project]:
        projects = database.get_all_projects()
        if not projects:
            return [Project("Add a project")]
        return projects

    def __load_todays_total_time(self) -> int:
        return database.get_todays_total_time()

    @property
    def formatted_todays_total_time(self) -> str:
        return format_project_time(math.floor(self.todays_total_time))



from typing import List
from blessed import Terminal
import math


def print_xy(term: Terminal, x: int, y: int, value: str, flush=True, max_width: int = None,
             center=False, trim=False):
    """trim and center both require max_width to be given"""
    if max_width is not None:
        empty_len = max(max_width - term.length(value), 0)

        if trim and term.length(value) > max_width:
            length = term.length(value)
            eles = term.split_seqs(value)[::-1]
            dot_count = 0

            for i, ele in enumerate(eles):
                # Skip if element is a sequence
                if len(term.strip_seqs(ele)) == 0:
                    continue

                if length > max_width:
                    eles[i] = ""
                    length -= 1
                else:
                    eles[i] = "."
                    dot_count += 1

                    if dot_count == 2:
                        break

            value = "".join(eles[::-1])

        if center:
            empty = " " * (empty_len // 2)
            print(term.move_xy(x - max_width // 2, y) +
                  f" {empty}{value}{empty} ", end="", flush=flush)
        else:
            empty = " " * empty_len
            print(term.move_xy(x, y) +
                  f" {value}{empty} ", end="", flush=flush)
    else:
        print(term.move_xy(x, y) + value, end="", flush=flush)


def print_lines_xy(term: Terminal, x: int, y: int, lines: List[str], flush=True, max_width: int = None,
                   center=False, trim=False):
    for i, line in enumerate(lines):
        print_xy(term, x, y + i, line, flush, max_width, center, trim)



from distutils.core import setup
setup(
    name='stargazing',         # How you named your package folder (MyLib)
    packages=['stargazing'],   # Chose the same as "name"
    package_data={'': ['audio/*', 'config/*',
                       'data/*', 'logs/*', 'pomodoro/*', 'project/*', 'res/*', 'utils/*']},
    version='0.2',      # Start with a small number and increase it with every change you make
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    license='MIT',
    # Give a short description about your library
    description='A terminal user interface study/work app',
    author='mtu2',                   # Type in your name
    # author_email='',      # Type in your E-Mail
    # Provide either the link to your github or to your website
    url='https://github.com/mtu2/stargazing',
    # I explain this later on
    download_url='https://github.com/mtu2/stargazing/archive/refs/tags/v0.2.tar.gz',
    # Keywords that define your package best
    keywords=['terminal', 'python', 'study', 'work', 'pomodoro'],
    install_requires=[            # I get to this in a second
        'ansicon',
        'blessed',
        'jinxed',
        'pafy',
        'python-vlc',
        'six',
        'toml',
        'ujson',
        'wcwidth',
        'youtube-dl',
    ],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    entry_points={
        'console_scripts': [
            'stargazing=stargazing.main:run_stargazing',
        ],
    },
)



import contextlib
import io
import threading
from typing import Any, Callable, Iterable, Union


def null_fn() -> None:
    return None


def check_null_fn(fn: Union[Callable[[], None], None]) -> Callable[[], None]:
    if fn is None:
        return null_fn
    return fn

def check_iterable(s: Union[str, Iterable[str]]) -> None:
    if isinstance(s, str):
        return [s]
    return s

def start_daemon_thread(target, args: Union[Iterable[Any], None] = None) -> threading.Thread:
    if not args:
        args = []
        
    thread = threading.Thread(target=target, args=args)
    thread.setDaemon(True)
    thread.start()
    return thread
        
class NullIO(io.StringIO):
    def write(self, txt: str) -> None:
       pass
    
def silent_stderr(fn):
    """Decorator to silence stderr output of functions."""
    def silent_fn(*args, **kwargs):
        with contextlib.redirect_stderr(NullIO()):
            return fn(*args, **kwargs)
    return silent_fn
